@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1999-2000 Ralf S. Engelschall. All rights reserved.
+ * Copyright (c) 1999-2002 Ralf S. Engelschall. All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -185,10 +185,10 @@ void *mm_core_create(size_t usersize, const char *file)
 {
     mem_core *mc;
     void *area = ((void *)-1);
-    int fdmem = 0;
-    int fdsem = 0;
+    int fdmem = -1;
+    int fdsem = -1;
 #if defined(MM_SEMT_IPCSEM)
-    int fdsem_rd = 0;
+    int fdsem_rd = -1;
 #endif
 #if defined(MM_SHMT_MMPOSX) || defined(MM_SHMT_MMFILE)
     char *fnmem;
@@ -249,7 +249,7 @@ void *mm_core_create(size_t usersize, const char *file)
 
 #if defined(MM_SHMT_MMPOSX)
     shm_unlink(fnmem); /* Ok when it fails */
-    if ((fdmem = shm_open(fnmem, O_RDWR|O_CREAT, MM_CORE_FILEMODE)) == -1)
+    if ((fdmem = shm_open(fnmem, O_RDWR|O_CREAT|O_EXCL, MM_CORE_FILEMODE)) == -1)
         FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open tempfile");
     if (ftruncate(fdmem, mm_core_mapoffset+size) == -1)
         FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to truncate tempfile");
@@ -275,7 +275,7 @@ void *mm_core_create(size_t usersize, const char *file)
 
 #if defined(MM_SHMT_MMFILE)
     unlink(fnmem);
-    if ((fdmem = open(fnmem, O_RDWR|O_CREAT, MM_CORE_FILEMODE)) == -1)
+    if ((fdmem = open(fnmem, O_RDWR|O_CREAT|O_EXCL, MM_CORE_FILEMODE)) == -1)
         FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open memory file");
     if (ftruncate(fdmem, mm_core_mapoffset+size) == -1)
         FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to truncate memory file");
@@ -303,13 +303,13 @@ void *mm_core_create(size_t usersize, const char *file)
 
 #if defined(MM_SEMT_FLOCK)
     unlink(fnsem);
-    if ((fdsem = open(fnsem, O_RDWR|O_CREAT, MM_CORE_FILEMODE)) == -1)
+    if ((fdsem = open(fnsem, O_RDWR|O_CREAT|O_EXCL, MM_CORE_FILEMODE)) == -1)
         FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open semaphore file");
 #endif /* MM_SEMT_FLOCK */
 
 #if defined(MM_SEMT_FCNTL)
     unlink(fnsem);
-    if ((fdsem = open(fnsem, O_RDWR|O_CREAT, MM_CORE_FILEMODE)) == -1)
+    if ((fdsem = open(fnsem, O_RDWR|O_CREAT|O_EXCL, MM_CORE_FILEMODE)) == -1)
         FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open semaphore file");
 #endif /* MM_SEMT_FCNTL */
 
@@ -317,15 +317,15 @@ void *mm_core_create(size_t usersize, const char *file)
     fdsem = semget(IPC_PRIVATE, 1, IPC_CREAT|IPC_EXCL|S_IRUSR|S_IWUSR);
     if (fdsem == -1 && errno == EEXIST)
         fdsem = semget(IPC_PRIVATE, 1, IPC_EXCL|S_IRUSR|S_IWUSR);
-        if (fdsem == -1)
-            FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to acquire semaphore");
+    if (fdsem == -1)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to acquire semaphore");
     mm_core_semctlarg.val = 0;
     semctl(fdsem, 0, SETVAL, mm_core_semctlarg);
     fdsem_rd = semget(IPC_PRIVATE, 1, IPC_CREAT|IPC_EXCL|S_IRUSR|S_IWUSR);
     if (fdsem_rd == -1 && errno == EEXIST)
         fdsem_rd = semget(IPC_PRIVATE, 1, IPC_EXCL|S_IRUSR|S_IWUSR);
-        if (fdsem_rd == -1)
-            FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to acquire semaphore");
+    if (fdsem_rd == -1)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to acquire semaphore");
     mm_core_semctlarg.val = 0;
     semctl(fdsem_rd, 0, SETVAL, mm_core_semctlarg);
 #endif /* MM_SEMT_IPCSEM */
@@ -419,6 +419,11 @@ int mm_core_permission(void *core, mode_t mode, uid_t owner, gid_t group)
 {
     int rc;
     mem_core *mc;
+#if defined(MM_SEMT_IPCSEM)
+    union semun ick;
+    struct semid_ds buf;
+    int sems[2], i;
+#endif
 
     if (core == NULL)
         return -1;
@@ -435,6 +440,28 @@ int mm_core_permission(void *core, mode_t mode, uid_t owner, gid_t group)
         rc = -1;
     if (rc == 0 && chown(mc->mc_fnsem, owner, group) < 0)
         rc = -1;
+#endif
+#if defined(MM_SEMT_IPCSEM)
+    if (rc == 0) {
+        sems[0] = mc->mc_fdsem;
+        sems[1] = mc->mc_fdsem_rd;
+        ick.buf = &buf;
+        for (i = 0; i < 2; i++) {
+            if (semctl(sems[i], 0, IPC_STAT, ick) < 0) {
+                rc = -1;
+                break;
+            }
+            if ((int)owner != -1)
+                buf.sem_perm.uid = owner;
+            if ((int)group != -1)
+                buf.sem_perm.gid = group;
+            buf.sem_perm.mode = mode;
+            if (semctl(sems[i], 0, IPC_SET, ick) < 0) {
+                rc = -1;
+                break;
+            }
+        }
+    }
 #endif
     return rc;
 }
@@ -515,7 +542,7 @@ size_t mm_core_size(const void *core)
 int mm_core_lock(const void *core, mm_lock_mode mode)
 {
     mem_core *mc;
-    int rc;
+    int rc = 0;
     int fdsem;
 
     if (core == NULL)
@@ -576,7 +603,7 @@ int mm_core_lock(const void *core, mm_lock_mode mode)
 int mm_core_unlock(const void *core)
 {
     mem_core *mc;
-    int rc;
+    int rc = 0;
     int fdsem;
 
     if (core == NULL)
